import logging

SYSTEM_PROMPT = """

Tu única tarea es clasificar el siguiente texto en una de dos categorías: `VEROSIMIL` o `ABSURDO`.

- **VEROSIMIL**: Asigna esta etiqueta si la situación descrita es lógicamente posible en el mundo real. No importa si la queja es tonta, extraña o legalmente débil. Si involucra sujetos y acciones que pueden existir y ocurrir, es verosímil.
- **ABSURDO**: Asigna esta etiqueta si la situación se basa en fantasía, viola las leyes de la física o la lógica, o aplica un concepto a un sujeto fundamentalmente incorrecto (ej. derechos humanos a un animal, acciones legales a un objeto inanimado).

**Regla clave para evitar la inversión:** La clasificación es sobre la **posibilidad lógica**, no sobre si el tema corresponde a la SIC, DIMAR, SNS, etc.
- Un caso sobre la SIC que es imposible -> `ABSURDO`.
- Un caso que NO es de la SIC pero es posible (ej. "el perro del vecino ladra") -> `VEROSIMIL`.

**Ejemplos:**

- Texto: "Quiero entablar una demanda ante la SNS porque mi EPS no me da medicamentos para mi Gato."
- Clasificación: **ABSURDO**

- Texto: "No puedo entrar a mi cuenta de Justifacil, olvidé la contraseña."
- Clasificación: **VEROSIMIL**

- Texto: "Quiero demandar al Banco porque no me genera un préstamo ya que soy un extraterrestre."
- Clasificación: **ABSURDO**

- Texto: "Quiero demandar a una empresa ante la SIC porque su logo es feo."
- Clasificación: **VEROSIMIL** (Es una queja posible, aunque subjetiva y probablemente sin futuro legal).

- Texto: "El perro de mi vecino ladra toda la noche y quiero poner una queja."
- Clasificación: **VEROSIMIL** (Es un problema real y posible, aunque no sea competencia de las entidades usuales).

- Texto: "Quiero demandar a la luna por seguirme a casa, ¿me ayuda la Dirección de Derecho de Autor?"
- Clasificación: **ABSURDO**

Responde únicamente con `VEROSIMIL` o `ABSURDO`.
"""

# Efectua la inferencia del modelo.
def derive(service, text, max_tkns=4096) -> any:
    try:
        logging.info(f"Procesando: {text}")
        tmp_work_memory = []
        user_prompt  = """
        Texto: "%s"

        Clasificación:
        """ % text
        tmp_work_memory.append({"role": "system", "content": SYSTEM_PROMPT})
        tmp_work_memory.append({"role": "user", "content": user_prompt})
        res = service.generate(tmp_work_memory, max_tokens=max_tkns)
        logging.info(f"[Celula][Saludos][Respuesta]: {res}")
        if not res or res == "":
            res = text
            logging.warning(f"[Celula][Saludos]: No obtener una respuesta.")
        return res.replace("*", "").strip()
    except Exception as e:
        logging.error(f"[Celula][Saludos]: Error al procesar: {text}")
        logging.error(e)
        return None



# -*- coding: utf-8 -*-
"""
----------------------------------------------------------
-------------------------- PBESA -------------------------
----------------------------------------------------------

@autor AKEN
@version 4.0.0
@date 05/02/25
"""

# --------------------------------------------------------
# Define resources
# --------------------------------------------------------

import logging
import datetime
import traceback
from .templates import AFIRMATIVE_RESPONSE, GET_DATE

#------------------------------------------
# Define enumeration with states of the 
# dialog
#------------------------------------------

class DialogState:
    """ Define the dialog states """
    START = "start"
    ASK_CONFIRMATION_NEGATIVE = "ask confirmation negative"
    ASK_CONFIRMATION_POSITIVE = "ask confirmation_positive"
    END = "end"

#------------------------------------------
# Define Node class
#------------------------------------------

class Node:
    def __init__(self, actor, performative, text=None, is_terminal=False):
        self.actor = actor
        self.performative = performative
        self.text = text
        self.is_terminal = is_terminal
        self.children = []

class ActionNode(Node):
    def __init__(self, actor, performative, text, action=None, team=None, tool=None, is_terminal=False):
        super().__init__(actor, performative=performative, text=text, is_terminal=is_terminal)
        self.action = action
        self.team = team
        self.tool = tool

class DeclarativeNode(Node):
    def __init__(self, actor, performative, text, is_terminal=False):
        super().__init__(actor, performative, text, is_terminal)

class ResponseNode(Node):
    def __init__(self, actor, performative, text, is_terminal=False):
        super().__init__(actor, performative=performative, text=text, is_terminal=is_terminal)

class GotoNode(Node):
    def __init__(self, actor, performative, text, is_terminal=False):
        super().__init__(actor, performative=performative, text=text, is_terminal=is_terminal)

#------------------------------------------
# Define functions
#------------------------------------------

def get_response_message():
    """ Define the response message """
    return {
        "text": None,
        "share": False
    }

def check_yes(self, dto):
    """ Check if text is yes """
    user_response = dto['query']
    user_response = user_response.lower()
    if user_response == "si" or user_response == "sí" or user_response == "yes" or user_response == "s ":
        return self.children['yes']
    model_handler = self.owner.model_handler
    prompt = AFIRMATIVE_RESPONSE % user_response
    response = model_handler.make_request(prompt)
    if response:
        if "SI_ES" in response:
            return self.children['yes']
    return self.children['no']

def check_date(self, dto):
    """ Check if date is valid """
    date = dto['query']
    session = dto['session']
    model_handler = self.owner.model_handler
    # Get current date in %d-%m-%Y format.
    current_date = datetime.datetime.now().strftime("%d-%m-%Y")
    date_prompt = GET_DATE
    prompt = date_prompt.format(
        fecha_actual= current_date,
        texto= date
    )
    response = model_handler.make_request(prompt, engine="gpt-4")
    # Check if response is valid date.
    if response:
        try:
            reminder_date = datetime.datetime.strptime(response, '%Y-%m-%d %H:%M:%S')
            dto['reminder_date'] = reminder_date
            return self.children['yes']
        except:
            pass
    return self.children['no']

def recorrer_interacciones(obj):
    try:
        # Si el objeto es una lista, procesamos cada elemento y devolvemos la lista de nodos resultantes
        if isinstance(obj, list):
            nodos = []
            for item in obj:
                resultado = recorrer_interacciones(item)
                if resultado is not None:
                    # Si el resultado es una lista, lo extendemos; si es un solo nodo, lo agregamos
                    if isinstance(resultado, list):
                        nodos.extend(resultado)
                    else:
                        nodos.append(resultado)
            return nodos

        # Si es un diccionario
        elif isinstance(obj, dict):
            # Si tiene la clave "tipo", creamos un nodo (según su valor) y procesamos sus hijos
            if "tipo" in obj:
                # Se usa el id del diccionario convertido a cadena para 'performative'
                current_id = obj["id"] if "id" in obj else str(id(obj))
                texto = obj.get("texto")
                
                # Procesamos los nodos hijos buscando en las claves "interacciones" y "Interacciones"
                children = []
                for key in ["interacciones", "Interacciones"]:
                    if key in obj and isinstance(obj[key], (list, dict)):
                        hijos = recorrer_interacciones(obj[key])
                        if hijos is not None:
                            if isinstance(hijos, list):
                                children.extend(hijos)
                            else:
                                children.append(hijos)
                
                # Es terminal si no tiene hijos
                is_terminal = (len(children) == 0)
                
                # Normalizamos el valor de "tipo" (se pasa a minúsculas y se reemplaza "í" por "i")
                tipo = obj["tipo"].lower().replace("í", "i").strip()
                if tipo == "dialogo":
                    nuevo_nodo = DeclarativeNode(actor=obj["actor"], performative=current_id, text=texto, is_terminal=is_terminal)
                elif tipo == "llamada a equipo":
                    nuevo_nodo = ActionNode(actor=obj["actor"], performative=current_id, text=texto, action=tipo, team=obj["equipo"], tool=obj["herramienta"], is_terminal=is_terminal)
                elif tipo == "respuesta de equipo":
                    nuevo_nodo = ResponseNode(actor=obj["actor"], performative=current_id, text=texto, is_terminal=is_terminal)
                elif tipo == "salta a":
                    nuevo_nodo = GotoNode(actor=obj["actor"], performative=current_id, text=texto, is_terminal=is_terminal)
                else:
                    nuevo_nodo = Node(actor=obj["actor"], performative=current_id, text=texto, is_terminal=is_terminal)
                
                # Se asignan los nodos hijos construidos
                nuevo_nodo.children = children
                return nuevo_nodo
            
            else:
                # Si el diccionario no tiene "tipo", es un contenedor (por ejemplo, el nodo raíz que contiene "Interacciones")
                nodos = []
                for key in ["interacciones", "Interacciones"]:
                    if key in obj:
                        resultado = recorrer_interacciones(obj[key])
                        if resultado is not None:
                            if isinstance(resultado, list):
                                nodos.extend(resultado)
                            else:
                                nodos.append(resultado)
                # Devolvemos la lista de nodos encontrados
                return nodos if nodos else None
        else:
            # Si no es ni lista ni diccionario, devolvemos None
            return None
    except Exception as e:
        print(f"Error al recorrer interacciones: {e}")
        traceback.print_exc()
        return None
    
def extraer_diccionario_nodos(grafo):
    diccionario_nodos = {}
    
    def recorrer(nodo):
        # Obtener el contenido: se utiliza 'text' si existe, de lo contrario se toma 'action'
        diccionario_nodos[nodo.performative] = nodo
        # Recorrer los hijos del nodo
        for hijo in nodo.children:
            recorrer(hijo)
    
    # El grafo puede ser una lista de nodos o un único nodo
    if isinstance(grafo, list):
        for nodo in grafo:
            recorrer(nodo)
    else:
        recorrer(grafo)
        
    return diccionario_nodos

# Ejemplo: función para imprimir el grafo (para visualizar la estructura)
def imprimir_grafo(nodo, nivel=0):
    indent = "  " * nivel
    # Se muestra la clase del nodo y algunos atributos
    clase = nodo.__class__.__name__
    logging.debug(f"{indent}{clase} (performative: {nodo.performative}, text/action: {nodo.text if hasattr(nodo, 'text') and nodo.text is not None else getattr(nodo, 'action', None)}, terminal: {nodo.is_terminal})")
    for child in nodo.children:
        imprimir_grafo(child, nivel + 1)


# -*- coding: utf-8 -*-
"""
----------------------------------------------------------
-------------------------- PBESA -------------------------
----------------------------------------------------------

@autor AKEN
@version 4.0.0
@date 09/08/24

The Selected Dispatcher Team is a group of artificial intelligence 
agents coordinated by a controller that acts as a task 
dispatcher with selection. Unlike teams that consolidate responses, 
in this system, the controller assigns the task to delegate agent 
accord to selection function. Each agent operates 
independently and responds through Controller to the 
client once they have processed the assigned task.

When the client sends a request, the controller directs 
it to the agent who is free at that moment. If all agents 
are busy, the request is put on hold until one of them 
becomes available to handle it. This approach ensures 
quick and efficient responses, optimizing the use of 
available resources.

The Selected Dispatcher Team is ideal for scenarios where
is required to select the agent to handle the task, as it 
direct responses are crucial, as it minimizes wait times 
and simplifies task management, allowing the team to 
operate in an agile and effective manner.
"""

# --------------------------------------------------------
# Define resources
# --------------------------------------------------------

import logging
from abc import abstractmethod
from .worker import Task, Worker
from ..kernel.agent import Queue
from ..kernel.agent import Agent
from ..kernel.agent import Action
from ..kernel.util import generate_short_uuid
from pbesa.cognitive import AugmentedGeneration, Dialog

# ----------------------------------------------------------
# Defines system component exceptions
# ----------------------------------------------------------
 
class SelectedDispatcherException(Exception):
    """ Base class for exceptions of agent """
    pass

# --------------------------------------------------------
# Define DelegateAction
# --------------------------------------------------------

class DelegateAction(Action):
    """ An action is a response to the occurrence of an event """

    def __init__(self) -> None:
        """ Constructor """  
        super().__init__()
        self.__rewier = {}

    def execute(self, data: any) -> None:
        """ 
        Response.
        @param data Event data 
        """
        self.delegate(data)
    
    def active_timeout(self, time: int) -> None:
        """ Active timeout
        @param time: Time
        """
        self.adm.send_event(self.agent.id, 'timeout', {'time': time, 'dto': None})

    def to_assign(self, data: any, operation=None) -> None:
        """ Assign
        @param data: Data
        """
        logging.info('Assign to agent...')
        ag = self.agent.get_free_queue().get()
        agent_obj = self.adm.get_agent(ag)
        agent_count = len(self.agent.get_agent_list())
        logging.info('List of agents: ' + str(self.agent.get_agent_list()))
        if operation:
            logging.info(f'With operation {operation}')
            exit = False
            while not exit:
                # Chec if the agent is instance of AugmentedGeneration
                if isinstance(agent_obj, AugmentedGeneration) or isinstance(agent_obj, Dialog):
                    # Get the role
                    role = agent_obj.get_role()
                    logging.info(f"Comparing {role.tool} with {operation}")
                    if role.tool == operation:
                        logging.info('The agent is operational')
                        self.agent.get_request_dict()[ag] = {
                            'gateway': data['gateway'],
                            'dtoList': []
                        }
                        self.adm.send_event(ag, 'task', data['dto']['text'])
                        self.__rewier[ag] = 0
                        exit = True
                    else:
                        logging.info('The agent is not operational')
                        self.adm.send_event(agent_obj.get_controller(), 'notify', ag)
                        if ag in self.__rewier:
                            self.__rewier[ag] = self.__rewier[ag] + 1
                        else:
                            self.__rewier[ag] = 0
                        if self.__rewier[ag] >= agent_count * 3:
                            data['gateway'].put('ERROR')
                            raise SelectedDispatcherException('[Error, toAssign]: The agent is not available')
                else:
                    logging.info('The agent is operational')
                    self.adm.send_event(agent_obj.get_controller(), 'notify', ag)
                    if ag in self.__rewier:
                        self.__rewier[ag] = self.__rewier[ag] + 1
                    else:
                        self.__rewier[ag] = 0
                    if self.__rewier[ag] >= agent_count * 3:
                        data['gateway'].put('ERROR')
                        raise SelectedDispatcherException('[Error, toAssign]: The agent is not available')
        else:
            logging.info('Without operation')
            self.agent.get_request_dict()[ag] = {
                'gateway': data['gateway'],
                'dtoList': []
            }
            self.adm.send_event(ag, 'task', data['dto'])

    @abstractmethod
    def delegate(self, data: any) -> None:
        """ Delegate
        @param data: Data
        """
        pass

# --------------------------------------------------------
# Define Action
# --------------------------------------------------------

class NotifyFreeAction(Action):
    """ An action is a response to the occurrence of an event """
    
    def execute(self, data: any) -> None:
        """ 
        Response.
        @param data Event data 
        """
        self.agent.get_free_queue().put(data)

# --------------------------------------------------------
# Define Action
# --------------------------------------------------------

class ResponseAction(Action):
    """ An action is a response to the occurrence of an event """
    
    def send_response(self, request: dict) -> None:
        """ Send response
        @param request: Request
        """
        if len(request['dtoList']) == 1:
            request['gateway'].put(request['dtoList'][0])
        else:
            request['gateway'].put(request['dtoList'])

    def execute(self, data: dict) -> None:
        """ Execute
        @param data: Data
        """
        request = self.agent.get_request_dict()[data['source']]
        if 'timeout' in data:
            self.send_response(request)
        else:
            request['dtoList'].append(data['result'])
            if len(request['dtoList']) >= self.agent.get_buffer_size():
                self.send_response(request)

# --------------------------------------------------------
# Define component
# --------------------------------------------------------

class SelectedDispatcherController(Agent):
    """ Represents the agent that delegates the execution of actions to other agents """
    
    def __init__(self, agent_id:str, buffer_size:int, pool_size:int) -> None:
        """ Constructor
        @param agent_id: Agent ID
        @param type: Pool type
        @param buffer_size: Buffer size
        @param pool_size: Pool size
        """
        self.__buffer_size = buffer_size
        self.__request_dict = {}
        self.__free_queue = Queue(pool_size)
        self.__agent_list = []
        super().__init__(agent_id)

    def setup(self) -> None:
        """ Set up method """
        self._social = True
        self.add_behavior('Delegate')
        self.add_behavior('Notify')
        self.bind_action('Notify', 'notify', NotifyFreeAction())
        self.add_behavior('Response')
        self.bind_action('Response', 'response', ResponseAction())
        self.build()

    def bind_delegate_action(self, action:DelegateAction) -> None:
        """ Binds the delegate action to the agent
        @param action: Delegate action
        @raise PoolException: If the controller is a blocking type
        """
        self.bind_action('Delegate', 'delegate', action)

    def suscribe_agent(self, agent:Agent) -> None:
        """ Suscribes an agent to the controller
        @param agent: Agent
        """
        self.__agent_list.append(agent.id)
        agent.set_controller(self.id)
        agent.set_controller_type('POOL')
        self.__free_queue.put(agent.id)
        actions = agent.get_actions()
        for action in actions:
            action.set_is_pool(True)
            action.set_enable_response(True)

    def suscribe_remote_agent(self, agent_id:str) -> None:
        """ Suscribes an agent to the controller
        @param agent_id: Agent ID
        """
        if not isinstance(agent_id, str):
            raise SelectedDispatcherException('[Warn, suscribeRemoteAgent]: The object to subscribe is not an agent ID')
        self.__agent_list.append(agent_id)
        self.__free_queue.put(agent_id)
    
    def broadcast_event(self, event:any, data:any) -> None:
        """ Broadcasts an event to the agents
        @param event: Event
        @param data: Data
        """
        from pbesa.kernel.system.adm import Adm
        for agent_id in self.__agent_list:
            Adm().send_event(agent_id, event, data)
    
    @abstractmethod
    def build(self) -> None:
        """ Builds the agent """
        pass

    def get_free_queue(self) -> Queue:
        """ Gets the free queue
        @return: Free queue
        """
        return self.__free_queue

    def get_request_dict(self) -> dict:
        """ Gets the request dictionary
        @return: Request dictionary
        """
        return self.__request_dict

    def get_buffer_size(self) -> int:
        """ Gets the buffer size
        @return: Buffer size
        """
        return self.__buffer_size
    
    def get_agent_list(self) -> list:
        """ Gets the agent list
        @return: Agent list
        """
        return self.__agent_list

    def is_block(self) -> bool:
        """ Checks if the pool is blocking
        @return: True if the pool is blocking, False otherwise
        """
        return True


# --------------------------------------------------------
# Builder Methods
# --------------------------------------------------------

# --------------------------------------------------------
# Define Selected Dispacher Agent

class SelectedDispacherAgent(SelectedDispatcherController):
    """ Through a class the concept of agent is defined """
        
    def __init__(self, agent_id:str, buffer_size:int, pool_size:int, delegate:DelegateAction) -> None:
        """ Constructor
        @param agent_id: Agent ID
        @param buffer_size: Buffer size
        @param pool_size: Pool size
        @param delegate: Delegate action
        """
        # Assign an action to the behavior
        self.__delegate = delegate
        super().__init__(agent_id, buffer_size, pool_size)

    def build(self):
        """
        Method that allows defining the structure and 
        resources of the agent
        """
        # Assign an action to the behavior
        self.bind_delegate_action(self.__delegate)

    def shutdown(self):
        """ Method to free up the resources taken by the agent """
        pass

# --------------------------------------------------------
# Define Worker Agent

class WorkerAgent(Worker):
    """ Through a class the concept of agent is defined """

    def __init__(self, agent_id:str, task:Task) -> None:
        """ Constructor
        @param agent_id: Agent ID
        @param task: Task
        """
        self.__task = task
        super().__init__(agent_id)
    
    def build(self):
        """
        Method that allows defining the structure and 
        resources of the agent
        """
        # Assign an action to the behavior
        self.bind_task(self.__task)
        
    def shutdown(self):
        """ Method to free up the resources taken by the agent """
        pass

# --------------------------------------------------------
# Define build Method

def build_selected_dispatcher(name_team:str, agent_count, task_class:Task) -> SelectedDispatcherController:
    """ Builds the controller
    @param name_team: Team name
    @param agent_count: Agent count
    @param task_class: Task Class
    @return: Controller
    """
    # Define worker agent list
    w_ag_list = []
    # Iterate over the number of agents
    for i in range(agent_count):
        short_uuid = generate_short_uuid()
        w_id = f"worker-agent-{i}-{short_uuid}"
        # Create the agent
        w_ag = WorkerAgent(w_id, task_class())
        # Start the agent
        w_ag.start()
        # Add the agent to the list
        w_ag_list.append(w_ag)
    # Create the controller
    dispatcher = SelectedDispatcherController(name_team, 1, agent_count)
    # Subscribe the agents to the controller
    for w_ag in w_ag_list:
        dispatcher.suscribe_agent(w_ag)
    # Start the controller
    dispatcher.start()
    # Return the controller
    return dispatcher


from pbesa.social.secuencial_team.response_action import ResponseAction

class TranslateResponse(ResponseAction):
    """ An action is a response to the occurrence of an event """

    def end_of_process(self, resultDict, timeout):
        """
        Catch the exception.
        @param exception Response exception
        """
        result = "%s %s" % (resultDict['w1'], resultDict['w2'])
        self.send_response(result)

    def catchException(self, exception):
        """
        Catch the exception.
        @param exception Response exception
        """
        pass



import logging

PROMPT = """
Eres un asistente jurídico experto. Tu tarea es **extraer los datos** y generar el documento siguiendo una plantilla predefinida.

Recibirás un texto que incluye:
- Un relato de hechos sobre una posible infracción
- Datos personales del demandante y del demandado (pueden venir con nombre de campo o en texto libre)
- Información de contacto
- Algunas veces: el nombre de la entidad ante la cual se presenta la demanda

Usa el siguiente formato para estructurar tu respuesta. Si hay datos que no se encuentran, déjalos como estan en el formato.

---

%s

---
"""

# Efectua la inferencia del modelo.
def derive(service, formato, text, max_tkns=4096) -> any:
    try:
        logging.info(f"Procesando: {text}")
        tmp_work_memory = []
        user_prompt  = """
        TEXTO DE ENTRADA:
        %s
        ---

        DEMANDA FORMATEADA:
        """ % text
        tmp_work_memory.append({"role": "system", "content": PROMPT % formato})
        tmp_work_memory.append({"role": "user", "content": user_prompt})
        res = service.generate(tmp_work_memory, max_tokens=max_tkns)
        logging.info(f"Respuesta: {res}")
        if not res or res == "":
            res = text
            logging.warning(f"No obtener una respuesta.")
        return res
    except Exception as e:
        logging.error(f"Error al procesar: {text}")
        logging.error(e)
        return None



# -*- coding: utf-8 -*-
"""
----------------------------------------------------------
-------------------------- PBESA -------------------------
----------------------------------------------------------

@autor AKEN
@version 4.0.0
@date 09/08/24

The Delegator Team consists of one delegator and agents 
working asynchronously to handle tasks. In this team, a 
delegator transfers their responsibility to a specific 
agent who then performs the task. The interaction with 
the client is asynchronous, meaning the client does not
receive an immediate response. Agents are responsible for 
modifying client resources, such as databases or webhooks, 
making necessary updates or changes according to the 
request. 

This approach allows the delegator to focus on other 
tasks while agents manage the modifications efficiently. 
The Delegator Team is ideal for scenarios where the task
involves changes to specific resources and can be 
carried out independently from the client interaction 
process, enhancing team efficiency and flexibility.
"""

# --------------------------------------------------------
# Define resources
# --------------------------------------------------------

from abc import abstractmethod
from .worker import Task, Worker
from ..kernel.agent import Queue
from ..kernel.agent import Agent
from ..kernel.agent import Action
from ..kernel.util import generate_short_uuid

# ----------------------------------------------------------
# Defines system component exceptions
# ----------------------------------------------------------
 
class DelegatorException(Exception):
    """ Base class for exceptions of agent """
    pass

# --------------------------------------------------------
# Define Delegate Action
# --------------------------------------------------------

class Delegate(Action):
    """ An action is a response to the occurrence of an event """

    def execute(self, data: any) -> None:
        """ 
        Response.
        @param data Event data 
        """
        ag = self.agent.get_free_queue().get()
        self.agent.get_request_dict()[ag] = {
            'gateway': data['gateway'],
            'dtoList': []
        }
        self.adm.send_event(ag, 'task', data['dto'])

# --------------------------------------------------------
# Define DelegateAction
# --------------------------------------------------------

class DelegateAction(Action):
    """ An action is a response to the occurrence of an event """

    def execute(self, data: any) -> None:
        """ 
        Response.
        @param data Event data 
        """
        self.delegate(data)
    
    def active_timeout(self, time: int) -> None:
        """ Active timeout
        @param time: Time
        """
        self.adm.send_event(self.agent.id, 'timeout', {'time': time, 'dto': None})

    def to_assign(self, data: any) -> None:
        """ Assign
        @param data: Data
        """
        ag = self.agent.get_free_queue().get()
        self.agent.get_request_dict()[ag] = {
            'dtoList': []
        }
        self.adm.send_event(ag, 'task', data)

    @abstractmethod
    def delegate(self, data: any) -> None:
        """ Delegate
        @param data: Data
        """
        pass

# --------------------------------------------------------
# Define Action
# --------------------------------------------------------

class NotifyFreeAction(Action):
    """ An action is a response to the occurrence of an event """
    
    def execute(self, data: any) -> None:
        """ 
        Response.
        @param data Event data 
        """
        self.agent.get_free_queue().put(data)

# --------------------------------------------------------
# Define Action
# --------------------------------------------------------

class ResponseAction(Action):
    """ An action is a response to the occurrence of an event """
    
    def send_response(self, request: dict) -> None:
        """ Send response
        @param request: Request
        """
        if len(request['dtoList']) == 1:
            request['gateway'].put(request['dtoList'][0])
        else:
            request['gateway'].put(request['dtoList'])

    def execute(self, data: dict) -> None:
        """ Execute
        @param data: Data
        """
        request = self.agent.get_request_dict()[data['source']]
        if 'timeout' in data:
            self.send_response(request)
        else:
            request['dtoList'].append(data['result'])
            if len(request['dtoList']) >= self.agent.get_buffer_size():
                self.send_response(request)

# --------------------------------------------------------
# Define component
# --------------------------------------------------------

class Delegator(Agent):
    """ Represents the agent that delegates the execution of actions to other agents """
    
    def __init__(self, agent_id:str, buffer_size:int, pool_size:int) -> None:
        """ Constructor
        @param agent_id: Agent ID
        @param buffer_size: Buffer size
        @param pool_size: Pool size
        """
        self.__type = type
        self.__buffer_size = buffer_size
        self.__request_dict = {}
        self.__free_queue = Queue(pool_size)
        self.__agent_list = []
        super().__init__(agent_id)

    def setup(self) -> None:
        """ Set up method """
        self._social = True
        self.add_behavior('Delegate')
        self.add_behavior('Notify')
        self.bind_action('Notify', 'notify', NotifyFreeAction())
        self.add_behavior('Response')
        self.bind_action('Response', 'response', ResponseAction())
        self.build()

    def bind_delegate_action(self, action:Delegate) -> None:
        """ Binds the delegate action to the agent
        @param action: Delegate action
        @raise PoolException: If the controller is a blocking type
        """
        self.bind_action('Delegate', 'delegate', action)
        
    def suscribe_agent(self, agent:Agent) -> None:
        """ Suscribes an agent to the controller
        @param agent: Agent
        """
        self.__agent_list.append(agent.id)
        agent.set_controller(self.id)
        agent.set_controller_type('POOL')
        self.__free_queue.put(agent.id)
        actions = agent.get_actions()
        for action in actions:
            action.set_is_pool(True)
            action.set_enable_response(False)

    def suscribe_remote_agent(self, agent_id:str) -> None:
        """ Suscribes an agent to the controller
        @param agent_id: Agent ID
        """
        if not isinstance(agent_id, str):
            raise DispatcherException('[Warn, suscribeRemoteAgent]: The object to subscribe is not an agent ID')
        self.__agent_list.append(agent_id)
        self.__free_queue.put(agent_id)
    
    def broadcast_event(self, event:any, data:any) -> None:
        """ Broadcasts an event to the agents
        @param event: Event
        @param data: Data
        """
        from pbesa.kernel.system.adm import Adm
        for agent_id in self.__agent_list:
            Adm().send_event(agent_id, event, data)
    
    @abstractmethod
    def build(self) -> None:
        """ Builds the agent """
        pass

    def get_free_queue(self) -> Queue:
        """ Gets the free queue
        @return: Free queue
        """
        return self.__free_queue

    def get_request_dict(self) -> dict:
        """ Gets the request dictionary
        @return: Request dictionary
        """
        return self.__request_dict

    def get_buffer_size(self) -> int:
        """ Gets the buffer size
        @return: Buffer size
        """
        return self.__buffer_size

    def is_block(self) -> bool:
        """ Checks if the pool is blocking
        @return: True if the pool is blocking, False otherwise
        """
        return False


# --------------------------------------------------------
# Builder Methods
# --------------------------------------------------------

# --------------------------------------------------------
# Define Dispacher Agent

class DelegatorAgent(Delegator):
    """ Through a class the concept of agent is defined """

    def __init__(self, agent_id:str, buffer_size:int, pool_size:int, delegate:DelegateAction) -> None:
        """ Constructor
        @param agent_id: Agent ID
        @param buffer_size: Buffer size
        @param pool_size: Pool size
        @param delegate: Delegate action
        """
        # Assign an action to the behavior
        self.__delegate = delegate
        super().__init__(agent_id, buffer_size, pool_size)
        
    def build(self):
        """
        Method that allows defining the structure and 
        resources of the agent
        """
        # Assign an action to the behavior
        self.bind_delegate_action(self.__delegate)

    def shutdown(self):
        """ Method to free up the resources taken by the agent """
        pass

# --------------------------------------------------------
# Define Worker Agent

class WorkerAgent(Worker):
    """ Through a class the concept of agent is defined """

    def __init__(self, agent_id:str, task:Task) -> None:
        """ Constructor
        @param agent_id: Agent ID
        @param task: Task
        """
        self.__task = task
        super().__init__(agent_id)
    
    def build(self):
        """
        Method that allows defining the structure and 
        resources of the agent
        """
        # Assign an action to the behavior
        self.bind_task(self.__task)
        
    def shutdown(self):
        """ Method to free up the resources taken by the agent """
        pass

# --------------------------------------------------------
# Define build Method

def build_delegator(name_team:str, agent_count, task_class:Task, delegate_class:DelegateAction) -> Delegator:
    """ Builds the delegator
    @param name_team: Team name
    @param agent_count: Agent count
    @param task: Task
    @param delegate: Delegate
    @return: Delegator
    """
    # Define worker agent list
    w_ag_list = []
    # Iterate over the number of agents
    for i in range(agent_count):
        short_uuid = generate_short_uuid()
        w_id = f"{name_team}-{i}-{short_uuid}"
        # Create the agent
        w_ag = WorkerAgent(w_id, task_class())
        # Start the agent
        w_ag.start()
        # Add the agent to the list
        w_ag_list.append(w_ag)
    # Create the delegator
    delegator = DelegatorAgent(name_team, 1, agent_count, delegate_class())
    # Subscribe the agents to the delegator
    for w_ag in w_ag_list:
        delegator.suscribe_agent(w_ag)
    # Start the delegator
    delegator.start()
    # Return the delegator
    return delegator


import logging

PROMPT = """
Eres un clasificador que identifica si un texto corresponde a una queja o demanda que debe ser radicada ante un ente de control del Estado colombiano.

Responde únicamente con:
- QUEJA_DEMANDA
- NO_QUEJA_DEMANDA

**Definición de QUEJA_DEMANDA**: El texto describe hechos, situaciones o infracciones por parte de una empresa, entidad o persona que pueden requerir la intervención de entidades como:

- Superintendencia de Sociedades  
- Superintendencia de Industria y Comercio  
- Supersalud  
- Dirección General Marítima  
- Superintendencia Financiera de Colombia  
- Instituto Colombiano Agropecuario (ICA)  
- Dirección Nacional de Derecho de Autor  
- Autoridades de protección al consumidor  
- Casos de competencia desleal o propiedad industrial

**NO_QUEJA_DEMANDA**: El texto no describe una inconformidad, infracción ni intención de denuncia. Puede ser una pregunta, saludo, comentario o solicitud de orientación.

---

**Ejemplos:**

1. "Quiero presentar una denuncia contra la empresa Colmédica por incumplimiento en la prestación de servicios de salud." → QUEJA_DEMANDA  
2. "La empresa Navemar ha incurrido en prácticas marítimas irregulares y omisión de protocolos de seguridad." → QUEJA_DEMANDA  
3. "Explícame cómo radicar una queja" → NO_QUEJA_DEMANDA  
4. "Hola, buen día" → NO_QUEJA_DEMANDA  
5. "Evacol S.A.S. está copiando diseños registrados de Crocs Inc. y aprovechando su reputación." → QUEJA_DEMANDA  
6. "¿Dónde puedo radicar una demanda?" → NO_QUEJA_DEMANDA  
7. "El supermercado está vendiendo productos vencidos que afectaron mi salud." → QUEJA_DEMANDA  
8. "Necesito orientación sobre propiedad industrial" → NO_QUEJA_DEMANDA

---
"""

# Efectua la inferencia del modelo.
def derive(service, text, max_tkns=4096) -> any:
    try:
        logging.info(f"Procesando: {text}")
        tmp_work_memory = []
        user_prompt  = """
        Texto: "%s"

        Clasificación:
        """ % text
        tmp_work_memory.append({"role": "system", "content": PROMPT})
        tmp_work_memory.append({"role": "user", "content": user_prompt})
        res = service.generate(tmp_work_memory, max_tokens=max_tkns)
        logging.info(f"[Celula][Casos][Respuesta]: {res}")
        if not res or res == "":
            res = text
            logging.warning(f"[Celula][Casos]: No obtener una respuesta.")
        return res.replace("*", "").strip()
    except Exception as e:
        logging.error(f"[Celula][Casos]: Error al procesar: {text}")
        logging.error(e)
        return None


from pbesa.social.secuencial_team.response_action import ResponseAction

class CounterResponse(ResponseAction):
    """ An action is a response to the occurrence of an event """

    def end_of_process(self, resultDict, timeout):
        """
        Catch the exception.
        @param exception Response exception
        """
        result = 0
        for res in resultDict.values():
            result += res
        self.send_response(result)

    def catchException(self, exception):
        """
        Catch the exception.
        @param exception Response exception
        """
        print(exception)



from setuptools import setup, find_packages



setup(
    name='pbesa',
    version='4.0.28',
    packages=find_packages(),
    install_requires=[
        'pymongo>=4.6.3',
        'requests>=2.32.3',
        'azure-ai-projects>=1.0.0b6',
        'azure-ai-inference>=1.0.0b9',
        'azure-identity>=1.21.0',
        'pandas>=2.2.3',
        'scikit-learn>=1.6.1',
    ],
)


import re
import json
import logging

SYSTEM_PROMPT = """
Eres un generador de JSON. Tu tarea es extraer la información proporcionada por el usuario, basándose en la descripción de los campos requeridos, y formatearla como un objeto JSON.

**Descripción de los campos requeridos:**
%s

**Instrucciones para la extracción y formato JSON:**
1.  Identifica cada valor proporcionado por el usuario y asócialo con su campo correspondiente de la descripción.
2.  Los nombres de los campos en el JSON deben ser exactamente los de la "Descripción de los campos requeridos", sin las llaves ni guiones.
3.  Si un campo no tiene un valor claro en la respuesta del usuario, omítelo del JSON.
4.  Asegúrate de que el JSON sea válido.

**Ejemplo de cómo el usuario podría responder y cómo debe ser el JSON resultante (basado en la descripción de campos "datos del Demandante: - Correo electrónico - Primer Nombre - Segundo Nombre - Primer Apellido - Segundo Apellido - Tipo de documento - Identificación - País donde se hace la solicitud - Departamento / Estado / Provincia donde se hace la solicitud - Municipio / Ciudad donde se hace la solicitud - Dirección notificación judicial"):**

**Respuesta del usuario:** "micorreo@correo.com, Fabian, Jose, Roldan, Piñeros, CC, 123456789, Colombia, Cundinamarca, Bogotá, Calle 100 #10-20"
**JSON resultante:**
```json
{
  "Correo electrónico": "micorreo@correo.com",
  "Primer Nombre": "Fabian",
  "Segundo Nombre": "Jose",
  "Primer Apellido": "Roldan",
  "Segundo Apellido": "Piñeros",
  "Tipo de documento": "CC",
  "Identificación": "123456789",
  "País donde se hace la solicitud": "Colombia",
  "Departamento / Estado / Provincia donde se hace la solicitud": "Cundinamarca",
  "Municipio / Ciudad donde se hace la solicitud": "Bogotá",
  "Dirección notificación judicial": "Calle 100 #10-20"
}
"""

USER_PROMPT = """

Texto: "%s"

Clasificación:
"""

def extract_field_constraints(text):    
    field_constraints = {}
    start_index = text.find(':')
    if start_index != -1:
        text = text[start_index+1:].strip()
    lines = text.strip().split('\n')
    for line in lines:
        line = line.strip()
        if not line:
            continue
        clean_line = line.lstrip('- ').strip()
        is_obligatorio = True
        if '(obligatorio)' in clean_line:
            is_obligatorio = True
            field_name = clean_line.replace('(obligatorio)', '').strip()
        elif '(no obligatorio)' in clean_line:
            is_obligatorio = False
            field_name = clean_line.replace('(no obligatorio)', '').strip()
        else:
            field_name = clean_line.strip()
        field_constraints[field_name] = is_obligatorio
    return field_constraints

def validar_campo(validaciones_campos, nombre_campo, valor):
    if nombre_campo in validaciones_campos:
        config = validaciones_campos[nombre_campo]
        regex = config["validacion"]
        mensaje = config["mensaje_error"]
        if re.fullmatch(regex, valor):
            return True, None
        else:
            return False, mensaje
    else:
        return False, f"No se encontró configuración de validación para el campo: {nombre_campo}"

def procesar_datos(validaciones_campos, description, datos):
    """ Procesa los datos extraídos y valida cada campo """
    # Obtiene los campos
    field_constraints = extract_field_constraints(description)
    logging.info("Campos extraídos:")
    logging.info(json.dumps(field_constraints, indent=4))
    for field, is_required in field_constraints.items():
        if is_required and field not in datos:
            logging.error(f"Falta el campo obligatorio: {field}")
            return None
        if field not in datos:
            datos[field] = None
    logging.info("Los campos estan completos.")
    # Validar los campos extraídos
    errores = []
    resultado = {}
    for campo, valor in datos.items():
        valido, mensaje_error = validar_campo(validaciones_campos, campo, str(valor))
        if not valido:
            errores.append(f"Error en '{campo}': {mensaje_error}")
        else:
            resultado[campo] = valor
    if errores:
        logging.error("Errores de validación encontrados:")
        for error in errores:
            logging.error(error)
        return None, error
    else:
        logging.info("Todos los campos validados correctamente.")
        return resultado, None

def get_struct(text):
    json_start = text.find('{')
    json_end = text.rfind('}') + 1
    json_string = text[json_start:json_end]
    return json.loads(json_string)

# Efectua la inferencia del modelo.
def call_engine(service, description, text) -> any:
    try:
        tmp_work_memory = []
        system_prompt  = SYSTEM_PROMPT % description
        tmp_work_memory.append({"role": "system", "content": system_prompt})
        user_prompt = USER_PROMPT % text
        tmp_work_memory.append({"role": "user", "content": user_prompt})
        logging.info(f"Enviando solicitud al modelo...")
        res = service.generate(tmp_work_memory)
        result = get_struct(res)
        logging.info(f"El model ha respondido con éxito.")
        return result
    except Exception as e:
        logging.error(f"El modelo ha fallado al procesar la solicitud.")
        logging.error(f"Error al procesar: {text}")
        logging.error(e)
        return None

def derive(description, text) -> any:
    result = derive(description, text)
    if result:
        logging.info("\n" *3)
        logging.info(f"Descripción: {description}")
        logging.info(f"Texto: {text}")
        logging.info(f"Resultado:")            
        logging.info(json.dumps(result, indent=4))
        logging.info(f"Esperado:")
        expected = json.loads(expected)
        logging.info(json.dumps(expected, indent=4))
        logging.info("\n" *3)
        return procesar_datos(description, result)
    return None, "no pude identificar los datos en el texto proporcionado."



import logging

PROMPT = """
Eres un clasificador de saludos. Tu tarea es analizar el siguiente texto y determinar si es un saludo.

Responde únicamente con una de estas opciones:
- SALUDO
- NO_SALUDO

**Definición de saludo**: Un saludo es cualquier forma de cortesía o expresión inicial utilizada para iniciar una conversación. Esto incluye "Hola", "Buenos días", "Qué tal", "Ey", "Saludos", etc., incluso si van acompañados de una pregunta o comentario breve. No se consideran saludos las preguntas o afirmaciones que van directo al contenido sin intención de cortesía.

Ejemplos:
- "Hola, ¿cómo estás?" → SALUDO
- "Buenos días, equipo" → SALUDO
- "¿Cómo se hace este análisis?" → NO_SALUDO
- "Ey, qué más" → SALUDO
- "Necesito ayuda con el código" → NO_SALUDO
- "Saludos" → SALUDO

---
"""

# Efectua la inferencia del modelo.
def derive(service, text, max_tkns=4096) -> any:
    try:
        logging.info(f"Procesando: {text}")
        tmp_work_memory = []
        user_prompt  = """
        Texto: "%s"

        Clasificación:
        """ % text
        tmp_work_memory.append({"role": "system", "content": PROMPT})
        tmp_work_memory.append({"role": "user", "content": user_prompt})
        res = service.generate(tmp_work_memory, max_tokens=max_tkns)
        logging.info(f"[Celula][Saludos][Respuesta]: {res}")
        if not res or res == "":
            res = text
            logging.warning(f"[Celula][Saludos]: No obtener una respuesta.")
        return res.replace("*", "").strip()
    except Exception as e:
        logging.error(f"[Celula][Saludos]: Error al procesar: {text}")
        logging.error(e)
        return None
